@@ -32,16 +32,19 @@ import re
 import logging
 import itertools
 import os
-import imp
 import html.parser
 import hashlib
 import json
+import yaml
+import sys
+import subprocess
+import io
 
 from enum import Enum
 from lxml import etree
 from lxml.cssselect import CSSSelector
 
-from .util import TrackSubClasses
+from .util import TrackSubClasses, import_module_from_source
 
 logger = logging.getLogger(__name__)
 
@@ -54,17 +57,19 @@ class FilterBase(object, metaclass=TrackSubClasses):
         self.job = job
         self.state = state
 
-    def _no_subfilters(self, subfilter):
-        if subfilter is not None:
-            raise ValueError('No subfilters supported for {}'.format(self.__kind__))
-
     @classmethod
     def filter_documentation(cls):
         result = []
         for sc in TrackSubClasses.sorted_by_kind(cls):
+            default_subfilter = getattr(sc, '__default_subfilter__', None)
             result.extend((
                 '  * %s - %s' % (sc.__kind__, sc.__doc__),
             ))
+            if hasattr(sc, '__supported_subfilters__'):
+                for key, doc in sc.__supported_subfilters__.items():
+                    result.append('      %s%s%s ... %s' % ('[' if key == default_subfilter else '', key,
+                                                           ']' if key == default_subfilter else '', doc))
+        result.append('\n[] ... Parameter can be supplied as unnamed value\n')
         return '\n'.join(result)
 
     @classmethod
@@ -81,18 +86,85 @@ class FilterBase(object, metaclass=TrackSubClasses):
 
         return data
 
+    @classmethod
+    def normalize_filter_list(cls, filter_spec):
+        for filter_kind, subfilter in cls._internal_normalize_filter_list(filter_spec):
+            filtercls = cls.__subclasses__.get(filter_kind, None)
+
+            if filtercls is None:
+                raise ValueError('Unknown filter kind: {} (subfilter {})'.format(filter_kind, subfilter))
+
+            if getattr(filtercls, '__no_subfilter__', False) and subfilter:
+                raise ValueError('No subfilters supported for {}'.format(filter_kind))
+
+            if isinstance(subfilter, dict) and hasattr(filtercls, '__supported_subfilters__'):
+                provided_keys = set(subfilter.keys())
+                allowed_keys = set(filtercls.__supported_subfilters__.keys())
+                unknown_keys = provided_keys.difference(allowed_keys)
+                if unknown_keys and '<any>' not in allowed_keys:
+                    raise ValueError('Filter "{}" does not support subfilter(s): {} (supported: {})'.format(filter_kind,
+                                                                                                            unknown_keys,
+                                                                                                            allowed_keys))
+
+            yield filter_kind, subfilter
+
+    @classmethod
+    def _internal_normalize_filter_list(cls, filter_spec):
+        if isinstance(filter_spec, str):
+            old_filter_spec = filter_spec
+
+            # Legacy string-based filter list specification:
+            # "filter1:param1,filter2,filter3,filter4:param4"
+            filter_spec = [dict([filter_kind.split(':', 1)]) if ':' in filter_kind else filter_kind
+                           for filter_kind in filter_spec.split(',')]
+
+            logger.warning('String-based filter definitions (%s) are deprecated, please convert to dict-style (see https://urlwatch.readthedocs.io/en/latest/deprecated.html):\n\n%s',
+                           old_filter_spec, yaml.dump(filter_spec, default_flow_style=False))
+
+        if isinstance(filter_spec, list):
+            for item in filter_spec:
+                if isinstance(item, str):
+                    filter_kind, subfilter = item, None
+                elif isinstance(item, dict):
+                    filter_kind, subfilter = next(iter(item.items()))
+
+                filtercls = cls.__subclasses__.get(filter_kind, None)
+
+                if isinstance(subfilter, dict):
+                    yield filter_kind, subfilter
+                elif subfilter is None:
+                    yield filter_kind, {}
+                elif hasattr(filtercls, '__default_subfilter__'):
+                    yield filter_kind, {getattr(filtercls, '__default_subfilter__'): subfilter}
+                else:
+                    yield filter_kind, subfilter
+
     @classmethod
     def process(cls, filter_kind, subfilter, state, data):
         logger.info('Applying filter %r, subfilter %r to %s', filter_kind, subfilter, state.job.get_location())
         filtercls = cls.__subclasses__.get(filter_kind, None)
-        if filtercls is None:
-            raise ValueError('Unknown filter kind: %s:%s' % (filter_kind, subfilter))
         return filtercls(state.job, state).filter(data, subfilter)
 
+    @classmethod
+    def filter_chain_needs_bytes(cls, filter):
+        # If the first filter is a bytes filter, return content in bytes instead of
+        # in unicode as that's what's required by the library used by that filter
+        first_filter = next(cls.normalize_filter_list(filter), None)
+        if first_filter is not None:
+            filter_kind, subfilter = first_filter
+            return cls.is_bytes_filter_kind(filter_kind)
+
+        return False
+
+    @classmethod
+    def is_bytes_filter_kind(cls, filter_kind):
+        return (filter_kind in [name for name, class_ in cls.__subclasses__.items()
+                                if getattr(class_, '__uses_bytes__', False)])
+
     def match(self):
         return False
 
-    def filter(self, data, subfilter=None):
+    def filter(self, data, subfilter):
         raise NotImplementedError()
 
 
@@ -137,21 +209,21 @@ class LegacyHooksPyFilter(FilterBase):
         self.hooks = None
         if os.path.exists(self.FILENAME):
             try:
-                self.hooks = imp.load_source('legacy_hooks', self.FILENAME)
+                self.hooks = import_module_from_source('legacy_hooks', self.FILENAME)
             except Exception as e:
                 logger.error('Could not load legacy hooks file: %s', e)
 
     def match(self):
         return self.hooks is not None
 
-    def filter(self, data, subfilter=None):
+    def filter(self, data):
         try:
             result = self.hooks.filter(self.job.get_location(), data)
             if result is None:
                 result = data
             return result
         except Exception as e:
-            logger.warn('Could not apply legacy hooks filter: %s', e)
+            logger.warning('Could not apply legacy hooks filter: %s', e)
             return data
 
 
@@ -160,21 +232,38 @@ class BeautifyFilter(FilterBase):
 
     __kind__ = 'beautify'
 
-    def filter(self, data, subfilter=None):
-        import jsbeautifier
-        import cssbeautifier
+    __no_subfilter__ = True
+
+    def filter(self, data, subfilter):
         from bs4 import BeautifulSoup as bs
         soup = bs(data, features="lxml")
-        scripts = soup.find_all('script')
-        for script in scripts:
-            if script.string is not None:
-                beautified_js = jsbeautifier.beautify(script.string)
-                script.string = beautified_js
-        styles = soup.find_all('style')
-        for style in styles:
-            if style.string is not None:
-                beautified_css = cssbeautifier.beautify(style.string)
-                style.string = beautified_css
+
+        try:
+            import jsbeautifier
+        except ImportError:
+            logger.info('"jsbeautifier" is not installed, will not beautify <script> tags')
+            jsbeautifier = None
+
+        if jsbeautifier is not None:
+            scripts = soup.find_all('script')
+            for script in scripts:
+                if script.string is not None:
+                    beautified_js = jsbeautifier.beautify(script.string)
+                    script.string = beautified_js
+
+        try:
+            import cssbeautifier
+        except ImportError:
+            logger.info('"cssbeautifier" is not installed, will not beautify <style> tags')
+            cssbeautifier = None
+
+        if cssbeautifier is not None:
+            styles = soup.find_all('style')
+            for style in styles:
+                if style.string is not None:
+                    beautified_css = cssbeautifier.beautify(style.string)
+                    style.string = beautified_css
+
         return soup.prettify()
 
 
@@ -183,19 +272,48 @@ class Html2TextFilter(FilterBase):
 
     __kind__ = 'html2text'
 
-    def filter(self, data, subfilter=None):
+    __supported_subfilters__ = {
+        'method': 'Method to use for conversion (default: re)',
+        '<any>': 'Method-specific options passed to html2text',
+    }
 
-        if subfilter is None:
+    __default_subfilter__ = 'method'
+
+    def filter(self, data, subfilter):
+        if 'method' in subfilter:
+            method = subfilter['method']
+            del subfilter['method']
+            options = subfilter
+        else:
             method = 're'
             options = {}
-        elif isinstance(subfilter, dict):
-            method = subfilter.pop('method')
-            options = subfilter
-        elif isinstance(subfilter, str):
-            method = subfilter
-            options = {}
+
         from .html2txt import html2text
-        return html2text(data, method=method, options=options)
+        return html2text(data, baseurl=getattr(self.job, 'url', getattr(self.job, 'navigate', '')),
+                         method=method, options=options)
+
+
+class Pdf2TextFilter(FilterBase):
+    """Convert PDF to plaintext"""
+    # Requires data to be in bytes (not unicode)
+    # Dependency: pdftotext (https://github.com/jalan/pdftotext), itself based
+    # on poppler (https://poppler.freedesktop.org/)
+    # Note: check pdftotext website for OS-specific dependencies for install
+
+    __kind__ = 'pdf2text'
+    __uses_bytes__ = True
+
+    __supported_subfilters__ = {
+        'password': 'PDF password for decryption',
+    }
+
+    def filter(self, data, subfilter):
+        # data must be bytes
+        if not isinstance(data, bytes):
+            raise ValueError('The pdf2text filter needs bytes input (is it the first filter?)')
+
+        import pdftotext
+        return '\n\n'.join(pdftotext.PDF(io.BytesIO(data), password=subfilter.get('password', '')))
 
 
 class Ical2TextFilter(FilterBase):
@@ -203,8 +321,9 @@ class Ical2TextFilter(FilterBase):
 
     __kind__ = 'ical2text'
 
-    def filter(self, data, subfilter=None):
-        self._no_subfilters(subfilter)
+    __no_subfilter__ = True
+
+    def filter(self, data, subfilter):
         from .ical2txt import ical2text
         return ical2text(data)
 
@@ -214,10 +333,14 @@ class JsonFormatFilter(FilterBase):
 
     __kind__ = 'format-json'
 
-    def filter(self, data, subfilter=None):
-        indentation = 4
-        if subfilter is not None:
-            indentation = int(subfilter)
+    __supported_subfilters__ = {
+        'indentation': 'Indentation level for pretty-printing',
+    }
+
+    __default_subfilter__ = 'indentation'
+
+    def filter(self, data, subfilter):
+        indentation = int(subfilter.get('indentation', 4))
         parsed_json = json.loads(data)
         return json.dumps(parsed_json, ensure_ascii=False, sort_keys=True, indent=indentation, separators=(',', ': '))
 
@@ -227,25 +350,37 @@ class GrepFilter(FilterBase):
 
     __kind__ = 'grep'
 
-    def filter(self, data, subfilter=None):
-        if subfilter is None:
+    __supported_subfilters__ = {
+        're': 'Lines matching this expression are kept (required)',
+    }
+
+    __default_subfilter__ = 're'
+
+    def filter(self, data, subfilter):
+        if 're' not in subfilter:
             raise ValueError('The grep filter needs a regular expression')
 
         return '\n'.join(line for line in data.splitlines()
-                         if re.search(subfilter, line) is not None)
+                         if re.search(subfilter['re'], line) is not None)
 
 
 class InverseGrepFilter(FilterBase):
-    """Filter which removes lines matching a regular expression"""
+    """Remove lines matching a regular expression"""
 
     __kind__ = 'grepi'
 
-    def filter(self, data, subfilter=None):
-        if subfilter is None:
+    __supported_subfilters__ = {
+        're': 'Lines matching this expression are removed (required)',
+    }
+
+    __default_subfilter__ = 're'
+
+    def filter(self, data, subfilter):
+        if 're' not in subfilter:
             raise ValueError('The inverse grep filter needs a regular expression')
 
         return '\n'.join(line for line in data.splitlines()
-                         if re.search(subfilter, line) is None)
+                         if re.search(subfilter['re'], line) is None)
 
 
 class StripFilter(FilterBase):
@@ -253,8 +388,9 @@ class StripFilter(FilterBase):
 
     __kind__ = 'strip'
 
-    def filter(self, data, subfilter=None):
-        self._no_subfilters(subfilter)
+    __no_subfilter__ = True
+
+    def filter(self, data, subfilter):
         return data.strip()
 
 
@@ -313,11 +449,17 @@ class GetElementById(FilterBase):
 
     __kind__ = 'element-by-id'
 
-    def filter(self, data, subfilter=None):
-        if subfilter is None:
+    __supported_subfilters__ = {
+        'id': 'ID of the element to filter for (required)',
+    }
+
+    __default_subfilter__ = 'id'
+
+    def filter(self, data, subfilter):
+        if 'id' not in subfilter:
             raise ValueError('Need an element ID for filtering')
 
-        element_by_id = ElementsBy(FilterBy.ATTRIBUTE, 'id', subfilter)
+        element_by_id = ElementsBy(FilterBy.ATTRIBUTE, 'id', subfilter['id'])
         element_by_id.feed(data)
         return element_by_id.get_html()
 
@@ -327,11 +469,17 @@ class GetElementByClass(FilterBase):
 
     __kind__ = 'element-by-class'
 
-    def filter(self, data, subfilter=None):
-        if subfilter is None:
+    __supported_subfilters__ = {
+        'class': 'HTML class attribute to filter for (required)',
+    }
+
+    __default_subfilter__ = 'class'
+
+    def filter(self, data, subfilter):
+        if 'class' not in subfilter:
             raise ValueError('Need an element class for filtering')
 
-        element_by_class = ElementsBy(FilterBy.ATTRIBUTE, 'class', subfilter)
+        element_by_class = ElementsBy(FilterBy.ATTRIBUTE, 'class', subfilter['class'])
         element_by_class.feed(data)
         return element_by_class.get_html()
 
@@ -341,11 +489,17 @@ class GetElementByStyle(FilterBase):
 
     __kind__ = 'element-by-style'
 
-    def filter(self, data, subfilter=None):
-        if subfilter is None:
+    __supported_subfilters__ = {
+        'style': 'HTML style attribute value to filter for (required)',
+    }
+
+    __default_subfilter__ = 'style'
+
+    def filter(self, data, subfilter):
+        if 'style' not in subfilter:
             raise ValueError('Need an element style for filtering')
 
-        element_by_style = ElementsBy(FilterBy.ATTRIBUTE, 'style', subfilter)
+        element_by_style = ElementsBy(FilterBy.ATTRIBUTE, 'style', subfilter['style'])
         element_by_style.feed(data)
         return element_by_style.get_html()
 
@@ -355,11 +509,17 @@ class GetElementByTag(FilterBase):
 
     __kind__ = 'element-by-tag'
 
-    def filter(self, data, subfilter=None):
-        if subfilter is None:
+    __supported_subfilters__ = {
+        'tag': 'HTML tag name to filter for (required)',
+    }
+
+    __default_subfilter__ = 'tag'
+
+    def filter(self, data, subfilter):
+        if 'tag' not in subfilter:
             raise ValueError('Need a tag for filtering')
 
-        element_by_tag = ElementsBy(FilterBy.TAG, subfilter)
+        element_by_tag = ElementsBy(FilterBy.TAG, subfilter['tag'])
         element_by_tag.feed(data)
         return element_by_tag.get_html()
 
@@ -369,8 +529,9 @@ class Sha1Filter(FilterBase):
 
     __kind__ = 'sha1sum'
 
-    def filter(self, data, subfilter=None):
-        self._no_subfilters(subfilter)
+    __no_subfilter__ = True
+
+    def filter(self, data, subfilter):
         sha = hashlib.sha1()
         sha.update(data.encode('utf-8', 'ignore'))
         return sha.hexdigest()
@@ -381,8 +542,9 @@ class HexdumpFilter(FilterBase):
 
     __kind__ = 'hexdump'
 
-    def filter(self, data, subfilter=None):
-        self._no_subfilters(subfilter)
+    __no_subfilter__ = True
+
+    def filter(self, data, subfilter):
         data = bytearray(data.encode('utf-8', 'ignore'))
         blocks = [data[i * 16:(i + 1) * 16] for i in range(int((len(data) + (16 - 1)) / 16))]
         return '\n'.join('%s  %s' % (' '.join('%02x' % c for c in block),
@@ -396,26 +558,18 @@ class LxmlParser:
 
     def __init__(self, filter_kind, subfilter, expr_key):
         self.filter_kind = filter_kind
-        if subfilter is None:
+        if expr_key not in subfilter:
             raise ValueError('Need %s for filtering' % (self.EXPR_NAMES[filter_kind],))
-        if isinstance(subfilter, str):
-            self.expression = subfilter
-            self.method = 'html'
-            self.exclude = None
-            self.namespaces = None
-        elif isinstance(subfilter, dict):
-            if expr_key not in subfilter:
-                raise ValueError('Need %s for filtering' % (self.EXPR_NAMES[filter_kind],))
-            self.expression = subfilter[expr_key]
-            self.method = subfilter.get('method', 'html')
-            self.exclude = subfilter.get('exclude')
-            self.namespaces = subfilter.get('namespaces')
-            if self.method not in ('html', 'xml'):
-                raise ValueError('%s method must be "html" or "xml", got %r' % (filter_kind, self.method))
-            if self.method == 'html' and self.namespaces is not None:
-                raise ValueError('Namespace prefixes only supported with "xml" method.')
-        else:
-            raise ValueError('%s subfilter must be a string or dict' % (filter_kind,))
+        self.expression = subfilter[expr_key]
+        self.method = subfilter.get('method', 'html')
+        self.exclude = subfilter.get('exclude')
+        self.namespaces = subfilter.get('namespaces')
+        self.skip = int(subfilter.get('skip', 0))
+        self.maxitems = int(subfilter.get('maxitems', 0))
+        if self.method not in ('html', 'xml'):
+            raise ValueError('%s method must be "html" or "xml", got %r' % (filter_kind, self.method))
+        if self.method == 'html' and self.namespaces is not None:
+            raise ValueError('Namespace prefixes only supported with "xml" method.')
         self.parser = (etree.HTMLParser if self.method == 'html' else etree.XMLParser)()
         self.data = ''
 
@@ -510,7 +664,21 @@ class LxmlParser:
         return [el for el in map(self._reevaluate, selected_elems) if el is not None]
 
     def get_filtered_data(self):
-        return '\n'.join(self._to_string(element) for element in self._get_filtered_elements())
+        elements = list(self._get_filtered_elements())
+        if self.skip:
+            elements = elements[self.skip:]
+        if self.maxitems:
+            elements = elements[:self.maxitems]
+        return '\n'.join(self._to_string(element) for element in elements)
+
+
+LXML_PARSER_COMMON_SUBFILTERS = {
+    'method': 'The method (html or xml) used for parsing',
+    'exclude': 'Elements to remove from the final result',
+    'namespaces': 'Mapping of XML namespaces for matching',
+    'skip': 'Number of elements to skip from the beginning (default: 0)',
+    'maxitems': 'Maximum number of items to return (default: all)',
+}
 
 
 class CssFilter(FilterBase):
@@ -518,7 +686,14 @@ class CssFilter(FilterBase):
 
     __kind__ = 'css'
 
-    def filter(self, data, subfilter=None):
+    __supported_subfilters__ = {
+        'selector': 'The CSS selector to use for filtering (required)',
+        **LXML_PARSER_COMMON_SUBFILTERS,
+    }
+
+    __default_subfilter__ = 'selector'
+
+    def filter(self, data, subfilter):
         lxml_parser = LxmlParser('css', subfilter, 'selector')
         lxml_parser.feed(data)
         return lxml_parser.get_filtered_data()
@@ -529,7 +704,14 @@ class XPathFilter(FilterBase):
 
     __kind__ = 'xpath'
 
-    def filter(self, data, subfilter=None):
+    __supported_subfilters__ = {
+        'path': 'The XPath to use for filtering (required)',
+        **LXML_PARSER_COMMON_SUBFILTERS,
+    }
+
+    __default_subfilter__ = 'path'
+
+    def filter(self, data, subfilter):
         lxml_parser = LxmlParser('xpath', subfilter, 'path')
         lxml_parser.feed(data)
         return lxml_parser.get_filtered_data()
@@ -540,25 +722,101 @@ class RegexSub(FilterBase):
 
     __kind__ = 're.sub'
 
-    def filter(self, data, subfilter=None):
-        if subfilter is None:
-            raise ValueError('{} needs a subfilter'.format(self.__kind__))
+    __supported_subfilters__ = {
+        'pattern': 'Regular expression to search for (required)',
+        'repl': 'Replacement string (default: empty string)',
+    }
 
-        # Allow for just specifying a regular expression (that will be removed)
-        if isinstance(subfilter, str):
-            subfilter = {'pattern': subfilter}
+    __default_subfilter__ = 'pattern'
+
+    def filter(self, data, subfilter):
+        if 'pattern' not in subfilter:
+            raise ValueError('{} needs a pattern'.format(self.__kind__))
 
         # Default: Replace with empty string if no "repl" value is set
-        return re.sub(subfilter.get('pattern'), subfilter.get('repl', ''), data)
+        return re.sub(subfilter['pattern'], subfilter.get('repl', ''), data)
 
 
 class SortFilter(FilterBase):
-    """Sort the results before comparison"""
+    """Sort input items"""
 
     __kind__ = 'sort'
 
-    def filter(self, data, subfilter=None):
-        data_list = data.splitlines()
-        data_list = sorted(data_list, key=str.casefold)
+    __supported_subfilters__ = {
+        'reverse': 'Set to true to reverse sorting order',
+        'separator': 'Item separator (default: newline)',
+    }
 
-        return '\n'.join(data_list)
+    __default_subfilter__ = 'separator'
+
+    def filter(self, data, subfilter):
+        reverse = (isinstance(subfilter, dict) and subfilter.get('reverse', False) is True)
+        separator = subfilter.get('separator', '\n')
+        return separator.join(sorted(data.split(separator), key=str.casefold, reverse=reverse))
+
+
+class ReverseFilter(FilterBase):
+    """Reverse input items"""
+
+    __kind__ = 'reverse'
+
+    __supported_subfilters__ = {
+        'separator': 'Item separator (default: newline)',
+    }
+
+    __default_subfilter__ = 'separator'
+
+    def filter(self, data, subfilter):
+        separator = subfilter.get('separator', '\n')
+        return separator.join(reversed(data.split(separator)))
+
+
+class ShellPipeFilter(FilterBase):
+    """Filter using a shell command"""
+
+    __kind__ = 'shellpipe'
+
+    __supported_subfilters__ = {
+        'command': 'Shell command to execute for filtering (required)',
+    }
+
+    __default_subfilter__ = 'command'
+
+    def filter(self, data, subfilter):
+        if 'command' not in subfilter:
+            raise ValueError('{} filter needs a command'.format(self.__kind__))
+
+        encoding = sys.getdefaultencoding()
+
+        # Work on a copy to not modify the outside environment
+        env = dict(os.environ)
+        env.update({
+            'URLWATCH_JOB_NAME': self.job.pretty_name() if self.job else '',
+            'URLWATCH_JOB_LOCATION': self.job.get_location() if self.job else '',
+        })
+
+        return subprocess.check_output(subfilter['command'], shell=True,
+                                       input=data.encode(encoding), env=env).decode(encoding)
+
+
+class OCRFilter(FilterBase):
+    """Convert text in images to plaintext using Tesseract OCR"""
+
+    __kind__ = 'ocr'
+    __uses_bytes__ = True
+
+    __supported_subfilters__ = {
+        'language': 'Language of the text (e.g. "fra" or "eng+fra")',
+        'timeout': 'Timeout (in seconds) for OCR (default 10 seconds)',
+    }
+
+    def filter(self, data, subfilter):
+        if not isinstance(data, bytes):
+            raise ValueError('The ocr filter needs bytes input (is it the first filter?)')
+
+        language = subfilter.get('language', None)
+        timeout = int(subfilter.get('timeout', 10))
+
+        import pytesseract
+        from PIL import Image
+        return pytesseract.image_to_string(Image.open(io.BytesIO(data)), lang=language, timeout=timeout)
