@@ -62,6 +62,23 @@ class JobState(object):
         self.error_ignored = False
         self._generated_diff = None
 
+    def __enter__(self):
+        try:
+            self.job.main_thread_enter()
+        except Exception as ex:
+            logger.info('Exception while creating resources for job: %r', self.job, exc_info=True)
+            self.exception = ex
+            self.traceback = traceback.format_exc()
+
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        try:
+            self.job.main_thread_exit()
+        except Exception as ex:
+            # We don't want exceptions from releasing resources to override job run results
+            logger.warning('Exception while releasing resources for job: %r', self.job, exc_info=True)
+
     def load(self):
         guid = self.job.get_guid()
         self.old_data, self.timestamp, self.tries, self.etag = self.cache_storage.load(self.job, guid)
@@ -79,6 +96,10 @@ class JobState(object):
 
     def process(self):
         logger.info('Processing: %s', self.job)
+
+        if self.exception:
+            return self
+
         try:
             try:
                 self.load()
@@ -115,6 +136,9 @@ class JobState(object):
     def get_diff(self):
         if self._generated_diff is None:
             self._generated_diff = self._generate_diff()
+            # Apply any specified diff filters
+            for filter_kind, subfilter in FilterBase.normalize_filter_list(self.job.diff_filter):
+                self._generated_diff = FilterBase.process(filter_kind, subfilter, self, self._generated_diff)
 
         return self._generated_diff
 
@@ -179,3 +203,9 @@ class Report(object):
         duration = (end - self.start)
 
         ReporterBase.submit_all(self, self.job_states, duration)
+
+    def finish_one(self, name):
+        end = datetime.datetime.now()
+        duration = (end - self.start)
+
+        ReporterBase.submit_one(name, self, self.job_states, duration)
