@@ -40,6 +40,7 @@ import urlwatch
 from requests.packages.urllib3.exceptions import InsecureRequestWarning
 
 from .util import TrackSubClasses
+from .filters import FilterBase
 
 requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
 
@@ -162,6 +163,14 @@ class JobBase(object, metaclass=TrackSubClasses):
     def retrieve(self, job_state):
         raise NotImplementedError()
 
+    def main_thread_enter(self):
+        """Called from the main thread before running the job"""
+        ...
+
+    def main_thread_exit(self):
+        """Called from the main thread after running the job"""
+        ...
+
     def format_error(self, exception, tb):
         return tb
 
@@ -171,7 +180,7 @@ class JobBase(object, metaclass=TrackSubClasses):
 
 class Job(JobBase):
     __required__ = ()
-    __optional__ = ('name', 'filter', 'max_tries', 'diff_tool', 'compared_versions')
+    __optional__ = ('name', 'filter', 'max_tries', 'diff_tool', 'compared_versions', 'diff_filter')
 
     # determine if hyperlink "a" tag is used in HtmlReporter
     LOCATION_IS_URL = False
@@ -198,6 +207,9 @@ class ShellJob(Job):
         if result != 0:
             raise ShellError(result)
 
+        if FilterBase.filter_chain_needs_bytes(self.filter):
+            return stdout_data
+
         return stdout_data.decode('utf-8')
 
 
@@ -284,6 +296,9 @@ class UrlJob(Job):
         # Save ETag from response into job_state, which will be saved in cache
         job_state.etag = response.headers.get('ETag')
 
+        if FilterBase.filter_chain_needs_bytes(self.filter):
+            return response.content
+
         # If we can't find the encoding in the headers, requests gets all
         # old-RFC-y and assumes ISO-8859-1 instead of UTF-8. Use the old
         # urlwatch behavior and try UTF-8 decoding first.
@@ -354,8 +369,12 @@ class BrowserJob(Job):
     def get_location(self):
         return self.navigate
 
+    def main_thread_enter(self):
+        from .browser import BrowserContext
+        self.ctx = BrowserContext()
+
+    def main_thread_exit(self):
+        self.ctx.close()
+
     def retrieve(self, job_state):
-        from requests_html import HTMLSession
-        session = HTMLSession()
-        response = session.get(self.navigate)
-        return response.html.html
+        return self.ctx.process(self.navigate)
