@@ -141,6 +141,16 @@ DEFAULT_CONFIG = {
             'to': '',
             'subject': '{count} changes: {jobs}'
         },
+        'ifttt': {
+            'enabled': False,
+            'key': '',
+            'event': '',
+        },
+        'xmpp': {
+            'enabled': False,
+            'sender': '',
+            'recipient': '',
+        },
     },
 
     'job_defaults': {
